/* gnome-rr.h: Display information gathering
 *
 * SPDX-FileCopyrightText: 2021  Emmanuele Bassi
 * SPDX-License-Identifier: LGPL-2.0-or-later
 */

#pragma once

#ifndef GNOME_DESKTOP_USE_UNSTABLE_API
# error GnomeRR is unstable API. You must define GNOME_DESKTOP_USE_UNSTABLE_API before including gnome-rr/gnome-rr.h
#endif

#include <gnome-rr/gnome-rr-config.h>
#include <gnome-rr/gnome-rr-output-info.h>
#include <gnome-rr/gnome-rr-screen.h>
